function [evecs, evals, A, B] = fem(shape, lb_param, max_num_evecs, fem_deg, A, B)
% shape is a struct with input fields:
% X, Y, Z, TRIV
% name is a string describing the shape
% ar is a scalar equal to the area of the shape (optional field)
%
% if A and B are not given as inputs they are computed in fem
% lb_param is either 'dir' or 'neu'

assert(strcmp(lb_param, 'dir') | strcmp(lb_param, 'neu'));

if(nargin == 3)
    fem_deg = 1;
end
if (nargin < 6)
    [A, B] = calc_A_B(shape, fem_deg);
end

%figure; imagesc(A); impixelinfo; title('A');
%figure; imagesc(B); impixelinfo; title('B');

[A, B, boundary_idx] = calc_dir_A_B(A, B, shape, lb_param);

num_evecs = min(size(A, 1) - 1, max_num_evecs);

% Adding B to A and subtracting 1 for numerical stability
rand('seed',0); randn('seed',0);
tic
[evecs, evals] = eigs(A + B, B, num_evecs, 'SM', struct('disp', 0));
evals = diag(evals) - 1;

%[evecs, evals] = eigs(A, B, num_evecs, 'SM', struct('disp',0));
%evals = diag(evals);

tc=toc; fprintf('done eigs(%6.2f sec)\n', tc);

assert(all(evals > -1e-6));

if strcmp(lb_param, 'dir')
    % assign zero to the boundary vertices
    num_vert = numel(shape.X);
    non_boundary_idx = setdiff(1:num_vert, boundary_idx);
    evecs_ = zeros(num_vert, num_evecs);
    evecs_(non_boundary_idx, :) = evecs;
    evecs = evecs_;
end

[evals, idx] = sort(evals);
evecs = evecs(:,idx);

if (0)
    % plot eigs limits
    plot_eig_limit(shape);
    plot_evecs(shape, [1, 4, 12, 18]);
end

if (0 && strcmp(shape.name, 'rec'))
    dbg_laplace_eigs(evals, evecs, shape);
end
end % function fem
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%5
function [] = plot_eig_limit(shape)

if(isfield(shape, 'ar'))
    k = length(evals);

    figure; plot(1:k, evals','-b.'); grid on; hold on; plot(1:k, 4*pi*[1:k]/shape.ar, 'r');
    xlabel('n');
    legend('\lambda_n', '4*\pi*n/ar')
    title('F.E.M');

    if(strcmp(shape.name, 'sphere'))
        for i=0:round(sqrt(max(evals(:))))
            plot([0, length(evals)], [i*(i+1), i*(i+1)], 'k');
        end;
    end
end
end % function plot_eigs_limits
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [A, B, boundary_idx] = calc_dir_A_B(A, B, shape, lb_param)

if (strcmp(lb_param, 'dir'))
    boundary_idx = calc_shape_boundary(shape);
    display(sprintf('num boundary vertices = %d / %d', length(boundary_idx), numel(shape.X)));
    %non_boundary_idx = setdiff(1:numel(shape.X), boundary_idx);

    A(boundary_idx, :) = []; A(:, boundary_idx) = [];

    B(boundary_idx, :) = []; B(:, boundary_idx) = [];
    
    if(0)
        plot_shape_holes(shape)
    end
else
    assert(strcmp(lb_param, 'neu'));
    boundary_idx = [];
end

end % function calc_dir_A_B