% Computes minimum distortion planar parametrization of a mesh
% 
% shape = parametrize(shape, N)
% 
% Input:
%  shape - a triangular mesh with the fields .TRIV, .X,.Y,.Z representing
%          the faces and the vertices of the mesh.
%  N     - number of landmarks to use
%  interptype - interpolation type; can be one of the following:
%   'landmark' standard LMDS interpolation, distorts holes
%   'project'  normal projection, low distortions
%
% Output:
%  fields .u,.v representing the parametrization coordinates of each vertex
%  are added to the shape structure. 
%
% (C) Alex Bronstein, Michael Bronstein, Niloy Mitra, 2009.
% All rights reserved.
function shape = parametrize(shape, N, interptype)

% Farthest point sampling
idx = [1];
DD = zeros(length(shape.X), N);
dmin = repmat(Inf, [length(shape.X) 1]);
str = ''; 
tic;
for k=2:N,
    D0 = repmat(Inf, [length(shape.X) 1]);
    D0(idx(end)) = 0;
    
    % TODO: this is extremely inefficient as the grid is re-initialized at
    % each iteration!
    D = fastmarch(shape.TRIV, shape.X, shape.Y, shape.Z, D0, struct('mode', 'single'));
    dmin = min(dmin,D(:));
    D(isinf(D)) = NaN;
    DD(:,k-1) = D;
    
    [dmax,i] = max(dmin);
    idx(k)= i;
    str = mprintf(str, 'Sampling: %d points  radius %f...', k, dmax);
end
D0 = repmat(Inf, [length(shape.X) 1]);
D0(idx(end)) = 0;
D = fastmarch(shape.TRIV, shape.X, shape.Y, shape.Z, D0, struct('mode', 'single'));
DD(:,end) = D;

mprintf(str, 'Sampled %d points with radius %f in %s\n', k, dmax, format_time(toc));

% Remove infinity distances if such appear
%%%%% MB - bred!!!  %i = find(~any(double(isinf(DD(idx,:)) | isnan(DD(idx,:))),2));
i = find(~any(double(isinf(DD(idx,:)) | isnan(DD(idx,:))),1));
idx = idx(i);
DD = DD(:,i);

% Embed landmarks into R^2 using MDS
shape.landmarks = idx;    
tic;
str = mprintf('', 'Computing embedding for %d landmarks...', length(idx));
D = DD(idx,:);    
D = 0.5*(D+D');
Y = mdscale(D,2); %X = mds(D, set_options('dim',2));
Y = bsxfun(@minus, Y, mean(Y ,1));

% Interpolate u,v coordinates to entire mesh
str = mprintf(str, 'Interpolating to %d points...', length(shape.X));
switch interptype,
    
    case 'landmark'
        dd = 0.5*bsxfun(@minus, mean(D.^2,2), (DD.^2)');
        u = Y \ dd;
        shape.u = u(1,:)';
        shape.v = u(2,:)';
        
    case 'project'
        %[DD,nearest] = sort(DD,2,'ascend');
        %DD = DD(:,1:3);
        %nearest = nearest(:,1:3);
        DD(isnan(DD)) = 0;
        nearest = zeros(size(DD,1), 3);
        for k=1:size(nearest,2),
            [d,i] = min(DD,[],2);
            nearest(:,k) = i;
            j = sub2ind(size(DD), [1:size(DD,1)]', i);
            DD(j) = NaN;
        end

        X = [shape.X(:) shape.Y(:) shape.Z(:)];
        X1 = X(idx(nearest(:,1)),:);
        X2 = X(idx(nearest(:,2)),:);
        X3 = X(idx(nearest(:,3)),:);

        P = cellfun(@(a,b,c)([a(:) b(:)]*inv([a(:) b(:)]'*[a(:) b(:)])*[a(:) b(:)]'), num2cell(X2-X1,2), num2cell(X3-X1,2), 'UniformOutput', false);
        x = cellfun(@(x,x0,P)(P*(x(:)-x0(:))+x0(:)), num2cell(X,2), num2cell(X1,2), P, 'UniformOutput', false);
        u = cellfun(@(x,a,b,c)([a(:)-c(:) b(:)-c(:)]\(x(:)-c(:))), num2cell(X,2), num2cell(X1,2), num2cell(X2,2), num2cell(X3,2), 'UniformOutput', false);
        u = cellfun(@(u)([u(:); 1-sum(u(:))]'), u, 'UniformOutput', false);
        u = cell2mat(u);

        % TODO: correct negative u's by selecting enclosing triangles

        % Interpolate parametrization coordinates
        shape.u = meshinterp(Y(:,1),u,nearest);
        shape.v = meshinterp(Y(:,2),u,nearest);
         
end

str = mprintf(str, 'Computed embedding for %d points in %s\n', length(shape.X), format_time(toc));



        

% return;
% 
% u = x; v = y;
% 
% 
